







def patronus(spell :Exception) -> str :

	"""
		Helper function that accepts Exception and provide pretty formatted message.
		P.S.: it must be "excepto_patronus", but in terms of conciseness it's just "patronus".
	"""

	return	f"{spell.__class__.__name__}: {spell}" if isinstance(spell, Exception) else str(spell)








def geminio(layer :type) -> type :

	"""
		Utility function that accepts Transmutable type class and declares a new class, that inherits from
		first one and forces it's self name to mimic the first one's, so the returned class is like a twin.
	"""

	class Transmute(layer):
		def __init__(self, *args, **kwargs):

			self.__class__.__name__ = layer.__name__
			super().__init__(*args, **kwargs)


	return	Transmute








def tentaclifors(*args, **kwargs) -> TypeError :

	"""
		Utility function that serves as a tool for ControlledTransmutation decorator to notify it's upper
		layer, or it's caller, about mutable chain initiation violation. That means every time in mutable
		chain initiation ControlledTransmutation class encounter not Transmutable class, this function will
		be returned by that ControlledTransmutation class, and as it is must be mandatory supplied with
		"_MUTABLE_CHAIN_HOOK" set to True, and it is not a class, it will be propagated to the very top
		layer, or caller, instead of used in mutable chain initiation.
	"""

	raise TypeError("Not Transmutable transmutation")

tentaclifors._MUTABLE_CHAIN_HOOK = True








def flagrate(counter :int | float | str) -> str :

	"""
		Utility function that returns either letter "s" or an empty string, depending on "counter".
		Basically this function upon calling with some quantity, allows to decide whether some
		string should be used in plural form or not. Argument "counter" must be of any numeric type,
		or in other words any value that if taken as an absolute value after float conversion will
		be legal for comparison with 1. All floating point "counter" values are plural by default.
		If "counter" is invalid, empty string will be returned.
	"""

	match counter:

		case float():	return "s"
		case int():		return "s" if abs(counter) != 1 else ""
		case str():

			try:	plural = abs(float(counter))
			except:	return str()
			else:	return "s" if "." in counter or plural != 1 else ""
		case _:		return str()








class Evanesco:

	""" Special magic object with spell that allows bypassing "loggy". """

	def vanishing(self, *args, **kwargs):	return NotImplemented
	def __getattribute__(self, attr :str):	return super().__getattribute__("vanishing")

evanesce = Evanesco()








def wingardium_leviosa(caller :object, attr :str, value :object, rebind =False):

	"""
		Utility function that accepts a Transmutable object "caller" and performs an escalation
		of a "value" with "attr" name right to the very top layer of a "caller". Any object other than
		Transmutable, which will not has an "_UPPER_LAYER_HOOK", will be just ignored.
	"""

	if	hasattr(caller, "_UPPER_LAYER_HOOK"):
		if	caller._UPPER_LAYER_HOOK:

			if(
				(attr == "loggy" and getattr(caller, "loggy") == evanesce or rebind)
				or
				(not hasattr(caller, attr) or rebind)

			):	setattr(caller, attr, value)
		else:	wingardium_leviosa(caller._UPPER_LAYER, attr, value, rebind)










import	os
import	unittest
from	pathlib								import Path
from	time								import sleep
from	shutil								import rmtree
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.tests.hagrid				import HagridTestCase
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.irma.shelve				import LibraryShelf
from	pygwarts.hagrid.planting.peeks		import BlindPeek








class BlindPeekCases(HagridTestCase):

	"""
		Ensures peeks algorithm works as intended
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.BLIND_PEEK_HANDLER): os.remove(cls.BLIND_PEEK_HANDLER)

		if	os.path.isdir(cls.INIT_SET_SPROUT):		rmtree(cls.INIT_SET_SPROUT)
		if	os.path.isdir(cls.INIT_SET_BOUGH_1):	rmtree(cls.INIT_SET_BOUGH_1)

	@classmethod
	def setUpClass(cls):

		cls.peekable1 = "peekable1"
		cls.make_loggy_file(cls, cls.BLIND_PEEK_HANDLER)
		if not os.path.isdir(cls.INIT_SET_BOUGH_1): os.makedirs(cls.INIT_SET_BOUGH_1)

		cls.peekable_sprout1	= os.path.join(cls.INIT_SET_SPROUT, "peekable1")
		cls.peekable_bough1		= os.path.join(cls.INIT_SET_BOUGH_1, "peekable1")
		cls.peekable_bough2		= os.path.join(cls.INIT_SET_BOUGH_2, "peekable1")
		cls.peekable_bough3		= os.path.join(cls.INIT_SET_BOUGH_3, "peekable1")




	def test_cached_blind_renew_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "cached_renew_1"
				init_level	= 10

			@BlindPeek("seeds")
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))


		with self.assertLogs("cached_renew_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_renew_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())


		with self.assertLogs("cached_renew_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_renew_1:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_renew_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_renew_1:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())








	def test_cached_blind_renew_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "cached_renew_2"
				init_level	= 10

			@BlindPeek("seeds")
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))

		bough_peek_mimic = first_peek -1
		self.test_case.seeds[self.peekable_sprout1] = bough_peek_mimic


		with self.assertLogs("cached_renew_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		with self.assertLogs("cached_renew_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_renew_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_renew_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertNotEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)








	def test_cached_blind_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "cached_1"
				init_level	= 10

			@BlindPeek("seeds", renew=False)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))


		with self.assertLogs("cached_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		with self.assertLogs("cached_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_1:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_1:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertNotEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)








	def test_cached_blind_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "cached_2"
				init_level	= 10

			@BlindPeek("seeds", renew=False)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))

		bough_peek_mimic = first_peek -1
		self.test_case.seeds[self.peekable_sprout1] = bough_peek_mimic


		with self.assertLogs("cached_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		with self.assertLogs("cached_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_2:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_2:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertNotEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)








	def test_cached_blind_picky_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "cached_picky_1"
				init_level	= 10

			@BlindPeek("seeds", renew=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)


		with self.assertLogs("cached_picky_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0.0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		with self.assertLogs("cached_picky_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0.0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_picky_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0.0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertNotEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)








	def test_cached_blind_picky_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "cached_picky_2"
				init_level	= 10

			@BlindPeek("seeds", renew=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)

		bough_peek_mimic = first_peek -1
		self.test_case.seeds[self.peekable_sprout1] = bough_peek_mimic


		with self.assertLogs("cached_picky_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		with self.assertLogs("cached_picky_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_picky_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertNotEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)








	def test_cached_blind_picky_renew_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "cached_picky_renew_1"
				init_level	= 10

			@BlindPeek("seeds", picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)


		with self.assertLogs("cached_picky_renew_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_picky_renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())


		with self.assertLogs("cached_picky_renew_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_1:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_picky_renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_picky_renew_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_1:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:cached_picky_renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())








	def test_cached_blind_picky_renew_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "cached_picky_renew_2"
				init_level	= 10

			@BlindPeek("seeds", picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)

		bough_peek_mimic = first_peek -1
		self.test_case.seeds[self.peekable_sprout1] = bough_peek_mimic


		with self.assertLogs("cached_picky_renew_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		with self.assertLogs("cached_picky_renew_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)


		with self.assertLogs("cached_picky_renew_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:Cached blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:cached_picky_renew_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertNotEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)








	def test_blind_renew_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "renew_1"
				init_level	= 10

			@BlindPeek("seeds", cache=False)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))


		with self.assertLogs("renew_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:renew_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())


		with self.assertLogs("renew_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:renew_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = int(os.path.getmtime(self.peekable_sprout1))
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("renew_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:renew_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())








	def test_blind_renew_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "renew_2"
				init_level	= 10

			@BlindPeek("seeds", cache=False)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))

		bough_peek_mimic = first_peek -1
		self.test_case.seeds[self.peekable_sprout1] = bough_peek_mimic


		with self.assertLogs("renew_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:renew_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:renew_2:Blind peek for \"{self.peekable_sprout1}\" old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		with self.assertLogs("renew_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:renew_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:renew_2:Blind peek for \"{self.peekable_sprout1}\" old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = int(os.path.getmtime(self.peekable_sprout1))
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("renew_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:renew_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:renew_2:Blind peek for \"{self.peekable_sprout1}\" old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertNotEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), second_peek)








	def test_blind_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "1"
				init_level	= 10

			@BlindPeek("seeds", cache=False, renew=False)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))


		with self.assertLogs("1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		with self.assertLogs("1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = int(os.path.getmtime(self.peekable_sprout1))
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), second_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)








	def test_blind_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "2"
				init_level	= 10

			@BlindPeek("seeds", cache=False, renew=False)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))

		bough_peek_mimic = first_peek -1
		self.test_case.seeds[self.peekable_sprout1] = bough_peek_mimic


		with self.assertLogs("2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:2:Blind peek for \"{self.peekable_sprout1}\" old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		with self.assertLogs("2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:2:Blind peek for \"{self.peekable_sprout1}\" old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), int(os.path.getmtime(self.peekable_sprout1))
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = int(os.path.getmtime(self.peekable_sprout1))
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:2:Blind peek for \"{self.peekable_sprout1}\" old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertNotEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), second_peek)








	def test_blind_picky_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "picky_1"
				init_level	= 10

			@BlindPeek("seeds", cache=False, renew=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)


		with self.assertLogs("picky_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:picky_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0.0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		with self.assertLogs("picky_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:picky_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0.0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = os.path.getmtime(self.peekable_sprout1)
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("picky_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:picky_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_1:Blind peek for \"{self.peekable_sprout1}\" old ring: 0.0",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), second_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)








	def test_blind_picky_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "picky_2"
				init_level	= 10

			@BlindPeek("seeds", cache=False, renew=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)

		bough_peek_mimic = first_peek -1
		self.test_case.seeds[self.peekable_sprout1] = bough_peek_mimic


		with self.assertLogs("picky_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:picky_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_2:Blind peek for \"{self.peekable_sprout1}\" old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		with self.assertLogs("picky_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:picky_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_2:Blind peek for \"{self.peekable_sprout1}\" old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = os.path.getmtime(self.peekable_sprout1)
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("picky_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:picky_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_2:Blind peek for \"{self.peekable_sprout1}\" old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertNotEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), second_peek)








	def test_blind_picky_renew_1(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "picky_renew_1"
				init_level	= 10

			@BlindPeek("seeds", cache=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)


		with self.assertLogs("picky_renew_1", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:picky_renew_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:picky_renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())


		with self.assertLogs("picky_renew_1", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:picky_renew_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:picky_renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = os.path.getmtime(self.peekable_sprout1)
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("picky_renew_1", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:picky_renew_1:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:picky_renew_1:Flourish stopped cause blind peek only for sprigs renew",
				case_loggy.output
			)
			self.assertNotIn(self.peekable_sprout1, self.test_case.seeds())








	def test_blind_picky_renew_2(self):

		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "picky_renew_2"
				init_level	= 10

			@BlindPeek("seeds", cache=False, picky=True)
			class grow(Transmutable):
				def __call__(*args):	pass

			class seeds(LibraryShelf):	pass


		self.test_case = Sakura()
		if	not os.path.isfile(self.peekable_sprout1):	self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)

		bough_peek_mimic = first_peek -1
		self.test_case.seeds[self.peekable_sprout1] = bough_peek_mimic


		with self.assertLogs("picky_renew_2", 10) as case_loggy:
			#################### grow to bough1
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		with self.assertLogs("picky_renew_2", 10) as case_loggy:
			#################### grow to bough2
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_2)
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{first_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(
				self.test_case.seeds(self.peekable_sprout1), os.path.getmtime(self.peekable_sprout1)
			)


		sleep(1.1)
		self.fmake(self.peekable_sprout1)
		second_peek = os.path.getmtime(self.peekable_sprout1)
		self.assertNotEqual(first_peek, second_peek)


		with self.assertLogs("picky_renew_2", 10) as case_loggy:
			#################### grow to bough3 (imitating source mtime change during Flourish)
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_3)
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:New blind peek \"{self.peekable_sprout1}\" new ring: "
				f"{second_peek}",
				case_loggy.output
			)
			self.assertIn(

				f"DEBUG:picky_renew_2:Blind peek for \"{self.peekable_sprout1}\" "
				f"old ring: {bough_peek_mimic}",
				case_loggy.output
			)
			self.assertIn(self.peekable_sprout1, self.test_case.seeds())
			self.assertNotEqual(self.test_case.seeds(self.peekable_sprout1), first_peek)
			self.assertEqual(self.test_case.seeds(self.peekable_sprout1), second_peek)








	def test_blind_comparators(self):

		self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))
		second_peek = first_peek -1

		comparisons = {

			"Flourish stopped by blind peek comparator":							lambda A,B : A <B,
			f"Blind peek for \"{self.peekable_sprout1}\" old ring: {second_peek}":	lambda A,B : B <A,
			f"Blind peek for \"{self.peekable_sprout1}\" old ring: {second_peek}":	lambda A,B : A <= B,
			f"Blind peek for \"{self.peekable_sprout1}\" old ring: {second_peek}":	lambda A,B : B <= A,
			"Flourish stopped by blind peek comparator":							lambda A,B : A == B,
			f"Blind peek for \"{self.peekable_sprout1}\" old ring: {second_peek}":	lambda A,B : A != B,
		}

		for i,( message,comparator ) in enumerate(comparisons.items()):
			class Sakura(Transmutable):
				class loggy(LibraryContrib):

					handler		= self.BLIND_PEEK_HANDLER
					init_name	= f"comparator-{i}"
					init_level	= 10

				@BlindPeek("seeds", renew=False, comparator=comparator)
				class grow(Transmutable):
					def __call__(*args):	pass

				class seeds(LibraryShelf):	pass

			self.test_case = Sakura()
			self.test_case.seeds[self.peekable_sprout1] = second_peek

			with self.assertLogs(f"comparator-{i}", 10) as case_loggy:

				self.test_case.grow(

					self.test_case,
					self.INIT_SET_SPROUT,
					Path(self.INIT_SET_SPROUT),
					Path(self.peekable_sprout1),
					Path(self.INIT_SET_BOUGH_1)
				)
				self.assertIn(f"DEBUG:comparator-{i}:{message}", case_loggy.output)
			self.test_case.loggy.close()








	def test_blind_picky_comparators(self):

		self.fmake(self.peekable_sprout1)
		first_peek = os.path.getmtime(self.peekable_sprout1)
		second_peek = first_peek -1

		comparisons = {

			"Flourish stopped by blind peek comparator":							lambda A,B : A <B,
			f"Blind peek for \"{self.peekable_sprout1}\" old ring: {second_peek}":	lambda A,B : B <A,
			f"Blind peek for \"{self.peekable_sprout1}\" old ring: {second_peek}":	lambda A,B : A <= B,
			f"Blind peek for \"{self.peekable_sprout1}\" old ring: {second_peek}":	lambda A,B : B <= A,
			"Flourish stopped by blind peek comparator":							lambda A,B : A == B,
			f"Blind peek for \"{self.peekable_sprout1}\" old ring: {second_peek}":	lambda A,B : A != B,
		}

		for i,( message,comparator ) in enumerate(comparisons.items()):
			class Sakura(Transmutable):
				class loggy(LibraryContrib):

					handler		= self.BLIND_PEEK_HANDLER
					init_name	= f"picky_comparator-{i}"
					init_level	= 10

				@BlindPeek("seeds", renew=False, picky=True, comparator=comparator)
				class grow(Transmutable):
					def __call__(*args):	pass

				class seeds(LibraryShelf):	pass

			self.test_case = Sakura()
			self.test_case.seeds[self.peekable_sprout1] = second_peek

			with self.assertLogs(f"picky_comparator-{i}", 10) as case_loggy:

				self.test_case.grow(

					self.test_case,
					self.INIT_SET_SPROUT,
					Path(self.INIT_SET_SPROUT),
					Path(self.peekable_sprout1),
					Path(self.INIT_SET_BOUGH_1)
				)
				self.assertIn(f"DEBUG:picky_comparator-{i}:{message}", case_loggy.output)
			self.test_case.loggy.close()








	def test_blind_extra_comparators(self):
		class Sakura(Transmutable):
			class loggy(LibraryContrib):

				handler		= self.BLIND_PEEK_HANDLER
				init_name	= "extra_comparator"
				init_level	= 10

			@BlindPeek("seeds", renew=False, comparator=lambda A,B : B <A or not os.path.isfile(B))
			class grow(Transmutable):
				def __call__(self, *args):	self.loggy.info("Proceed growing")

			class seeds(LibraryShelf):	pass

		self.fmake(self.peekable_sprout1)
		first_peek = int(os.path.getmtime(self.peekable_sprout1))

		self.test_case = Sakura()
		self.test_case.seeds[self.peekable_sprout1] = first_peek

		with self.assertLogs("extra_comparator", 10) as case_loggy:
			self.test_case.grow(

				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				Path(self.peekable_sprout1),
				Path(self.INIT_SET_BOUGH_1)
			)
		self.assertIn(

			f"DEBUG:extra_comparator:New blind peek \"{self.peekable_sprout1}\" new ring: "
			f"{first_peek}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:extra_comparator:Blind peek for \"{self.peekable_sprout1}\" old ring: {first_peek}",
			case_loggy.output
		)
		self.assertIn("INFO:extra_comparator:Proceed growing", case_loggy.output)








	def test_blind_raising_comparators(self):

		comparators = (

			lambda		: True,
			lambda A	: bool(A),
			lambda A	: A **2,
			lambda A,B	: A +B,
			lambda A,B	: "B <A",
			lambda A,B,C: A +B +C,
			lambda A,B,C: bool(A +B +C),
		)

		for comparator in comparators:

			try:

				class Sakura(Transmutable):

					@BlindPeek("seeds", comparator=comparator)
					class raising_grow(Transmutable):
						def __call__(*args):	pass


			except	Exception as E:

				self.assertIsInstance(E, TypeError)
				self.assertEqual(
					str(E), "Blind peek comparator must be Callable[[ int|float, int|float ], bool]"
				)








if	__name__ == "__main__" : unittest.main(verbosity=2)










from typing									import List
from typing									import Dict
from csv									import reader	as csv_reader
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.magical.chests				import KeyChest
from pygwarts.magical.spells				import patronus
from pygwarts.magical.spells				import flagrate
from pygwarts.filch.linkindor				import EUI48_format
from pygwarts.filch.nettherin				import validate_ip4








class MaraudersMap(KeyChest):

	"""
		filch utility class that serves as a runtime database for network traffic. Allows mapping of
		hosts addresses with corresponding information to be accessed as a dictionary. For current
		implementation following mapping can be maintained:
					MAC		(host physical address)
			IP4		NAME	(host alias/netbios name)
					DESC	(host description)
		and also
					IP4		(host IP version 4 address)
			MAC		NAME	(host alias/netbios name)
					DESC	(host description)
		According to scheme above, there will be two dictionaries in a KeyChest, IP4 and MAC. Every IP4
		and MAC addresses will be mapped with each other in corresponding dictionary, along with two
		additional fields. It is possible to have only IP4 or MAC for mapping, so it will be mapped with
		Nones and represents the addresses collection.
	"""

	@property
	def ip4(self) -> Dict[str,Dict[str,str]] | None :

		""" Property that returns a dictionary mapped to "IP4" or None. """

		if	isinstance(mapping := self["IP4"], dict): return mapping


	def ip4map(self, addr :str) -> Dict[str,str] | None :

		"""
			Mapping method that accepts IP4 address string and searches for corresponding mapping.
			If mapping found and it is a dictionary, returns it, no matter it's content.
			Returns None in all other cases.
		"""

		if	isinstance(self.ip4, dict) and isinstance(mapping := self.ip4.get(str(addr)), dict):
			return mapping


	def ip4map_mac(self, addr :str) -> str | None :

		"""
			Mapping method that accepts IP4 address string and searches for corresponding mapping.
			If mapping found and it is a dictionary that has "MAC" mapped to a string, returns it.
			Returns None in all other cases.
		"""

		if	isinstance(mapping := self.ip4map(addr), dict) and isinstance(mac := mapping.get("MAC"), str):
			return mac


	def ip4map_name(self, addr :str) -> str | None :

		"""
			Mapping method that accepts IP4 address string and searches for corresponding mapping.
			If mapping found and it is a dictionary that has "NAME" mapped to a string, returns it.
			Returns None in all other cases.
		"""

		if	isinstance(mapping := self.ip4map(addr), dict) and isinstance(name := mapping.get("NAME"), str):
			return name


	def ip4map_desc(self, addr :str) -> str | None :

		"""
			Mapping method that accepts IP4 address string and searches for corresponding mapping.
			If mapping found and it is a dictionary that has "DESC" mapped to a string, returns it.
			Returns None in all other cases.
		"""

		if	isinstance(mapping := self.ip4map(addr), dict) and isinstance(desc := mapping.get("DESC"), str):
			return desc




	@property
	def mac(self) -> Dict[str,Dict[str,str]] | None :

		""" Property that returns a dictionary mapped to "MAC" or None. """

		if	isinstance(mapping := self["MAC"], dict): return mapping


	def macmap(self, addr :str) -> Dict[str,str] | None :

		"""
			Mapping method that accepts physical address string and searches for corresponding mapping.
			If mapping found and it is a dictionary, returns it, no matter it's content.
			Returns None in all other cases.
		"""

		if	isinstance(self.mac, dict) and isinstance(mapping := self.mac.get(str(addr)), dict):
			return mapping


	def macmap_ip4(self, addr :str) -> str | None :

		"""
			Mapping method that accepts physical address string and searches for corresponding mapping.
			If mapping found and it is a dictionary that has "IP4" mapped to a string, returns it.
			Returns None in all other cases.
		"""

		if	isinstance(mapping := self.macmap(addr), dict) and isinstance(ip4 := mapping.get("IP4"), str):
			return ip4


	def macmap_name(self, addr :str) -> str | None :

		"""
			Mapping method that accepts physical address string and searches for corresponding mapping.
			If mapping found and it is a dictionary that has "NAME" mapped to a string, returns it.
			Returns None in all other cases.
		"""

		if	isinstance(mapping := self.macmap(addr), dict) and isinstance(name := mapping.get("NAME"), str):
			return name


	def macmap_desc(self, addr :str) -> str | None :

		"""
			Mapping method that accepts physical address string and searches for corresponding mapping.
			If mapping found and it is a dictionary that has "DESC" mapped to a string, returns it.
			Returns None in all other cases.
		"""

		if	isinstance(mapping := self.macmap(addr), dict) and isinstance(desc := mapping.get("DESC"), str):
			return desc








	class CSV(Transmutable):

		"""
			Helper class that serves as an instrument of maintaining the upper layer "database" by
			populating it from readable csv file. Must be supplied with two positional arguments
			"path" that represents path to csv file and "delimiter" to parse file. It also accepts
			keyword-only arguments that must represents the indices of corresponding fields in
			csv file lines. For every index provided, if it fits into csv file line, corresponding value
			will be extracted. For crucial IP4 and MAC fields corresponding value will be verified.
			If some value cannot be obtained it will be set to None. IP4 and MAC mapping will proceed
			only for validated values.
		"""

		def __call__(
						self,
						path		:str,
						delimiter	:str	=",",
						*,
						IP4			:int	=None,
						MAC			:int	=None,
						NAME		:int	=None,
						DESC		:int	=None,
					):

			self.loggy.debug(f"Reading csv from \"{path}\"")
			IP4_state = len(self._UPPER_LAYER["IP4"]) if self._UPPER_LAYER["IP4"] is not None else 0
			MAC_state = len(self._UPPER_LAYER["MAC"]) if self._UPPER_LAYER["MAC"] is not None else 0
			count = 0

			try:

				with open(path) as map_source:
					for row in csv_reader(map_source, delimiter=delimiter):


						current_IP4		= validate_ip4(self.extract(IP4, row))
						current_MAC		= EUI48_format(self.extract(MAC, row))
						current_NAME	= self.extract(NAME, row)
						current_DESC	= self.extract(DESC, row)
						count += 1


						if	current_IP4 is not None:

							self._UPPER_LAYER(

								current_IP4,
								{
									"MAC":	current_MAC,
									"NAME":	current_NAME,
									"DESC":	current_DESC,
								},
								"IP4",
								mapped=False,
							)
							self.loggy.debug(f"IP version 4 address {current_IP4} mapping done")


						if	current_MAC is not None:

							self._UPPER_LAYER(

								current_MAC,
								{
									"IP4":	current_IP4,
									"NAME":	current_NAME,
									"DESC":	current_DESC,
								},
								"MAC",
								mapped=False,
							)
							self.loggy.debug(f"Physical address {current_MAC} mapping done")


				self.loggy.debug(f"{count} record{flagrate(count)} read")


			except	Exception as E : self.loggy.debug(f"Failed to open CSV due to {patronus(E)}")
			else:

				if	self._UPPER_LAYER["IP4"] is not None:

					IP4_new = (ip4l := len(self._UPPER_LAYER["IP4"])) - IP4_state
					self.loggy.debug(f"{IP4_new} new IP4 mapping{flagrate(IP4_new)} done")
					self.loggy.debug(f"{ip4l} IP4 mapping{flagrate(ip4l)} total")


				if	self._UPPER_LAYER["MAC"] is not None:

					MAC_new = (macl := len(self._UPPER_LAYER["MAC"])) - MAC_state
					self.loggy.debug(f"{MAC_new} new MAC mapping{flagrate(MAC_new)} done")
					self.loggy.debug(f"{macl} MAC mapping{flagrate(macl)} total")




		def extract(self, index :int, source :List[str]) -> str | None :

			"""
				Utility method that will try to extract "index" value from "source" list and return
				it. Assumes "source" is a list of strings. Returns None if "source" is not a list or
				"index" is not an integer in range of "source" length.
			"""

			if	isinstance(source, list):
				if	isinstance(index, int):
					if	-1 <index <len(source):


						record = source[index]
						self.loggy.debug(f"Index {index} record \"{record}\" obtained")


						return	record
					else:		self.loggy.debug(f"Out of range index {index}")
				else:			self.loggy.debug(f"Invalid index {index}")
			else:				self.loggy.debug(f"Invalid source {source}")










from typing									import Any
from typing									import Dict
from typing									import Literal
from typing									import Optional
from typing									import Sequence
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.magical.spells				import flagrate
from pygwarts.magical.spells				import patronus
from pygwarts.hedwig.mail.builder			import LetterBuilder
from pygwarts.hedwig.mail.utils				import EmailValidator








class ObjectField(Transmutable):

	"""
		hedwig.mail core super class that is the base for a letter building. Represents the object
		model that will maintains some "field_value", which is obtainable via __call__, and has the
		ability to put itself in the upper layer LetterBuilder object, as a mapping of some special
		name for current "field_value" to itself. The optional field "modifiers" might be any type
		and might serve as some modifier for a "field_value" or just as some another value to be used
		somehow. Current super class __call__ defines the basic behavior for any field that will inherit,
		that it suggest the safety way to obtain "field_value". Any ObjectField child is supposed to
		maintain it's "field_value" the same way and implement the same interface. Any ObjectField
		child as a letter field might be represented multiple times, as "make_field" method handles.
	"""


	field_value	:Any
	modifiers	:Optional[Any]


	def __call__(self) -> Any :

		if	hasattr(self, "field_value"):
			self.loggy.debug(f"Fetched value of {type(self.field_value)}")

			return	self.field_value
		else:		self.loggy.error(f"Field {self} must have value")


	def make_field(self, field_name :str) -> Literal[True] | None :

		"""
			Core method for all letter's fields functionality, that implements mapping of fields
			for building the letter. Must be called with "field_name" string value that will be used
			as a key for mapping with the object itself. The current field object must be declared
			directly under LetterBuilder class, so this method will be able to do actual mapping in
			upper layer KeyChest. By default, if upper layer KeyChest does not content any mapping
			for "field_name" yet, actual mapping will be the list that will content current field
			object. Any further invocations of "make_field" for different field object, but with
			the same "field_name" will cause mapped list population. This allows multiple values
			to be obtained for a mapped "field_name" once the LetterBuilder object will build letter.
			Returns True if mapping was successfully made for current object, or None otherwise.
		"""

		if	isinstance(field_name, str):
			if	hasattr(self, "_UPPER_LAYER"):
				if	isinstance(self._UPPER_LAYER, LetterBuilder):


					match self._UPPER_LAYER[field_name]:

						case None:		self._UPPER_LAYER[field_name] = [ self ]
						case list():	self._UPPER_LAYER[field_name].append(self)
						case _:

							self.loggy.error(f"{self._UPPER_LAYER} {field_name} field is busy")
							return
					return	True


				else:	self.loggy.debug(f"Improper builder {self._UPPER_LAYER}")
			else:		self.loggy.debug("Builder not found")
		else:			self.loggy.debug(f"Field name \"{field_name}\" must be a string")








class StringField(ObjectField):

	"""
		hedwig.mail core super class, that is a child of ObjectField, that implements maintaining of only
		"field_value" of type string. In __call__ obtained "field_value" will be checked to be a string,
		so to be returned. Otherwise None will be returned with corresponding log message. This type of
		letter field is more subjective than ObjectField, but still general, as the original design of
		hedwig letter building does not assumes any substantive field in letter to be a simple string,
		but instead there are default fields that are constructed upon StringField:
			TextField and AttachmentField.
	"""

	field_value	:str

	def __call__(self) -> str | None :

		if	(current_value := super().__call__()) is not None:
			if	isinstance(current_value, str):

				flen = len(self.field_value)
				self.loggy.debug(f"Fetched {flen} symbol{flagrate(flen)}")


				return	self.field_value


			else:	self.loggy.error(f"Field {self} value must be string")
		else:		self.loggy.debug(f"No value fetched for field")








class TextField(StringField):

	"""
		hedwig.mail core class that is a child of StringField and is a super class for key-word formatted
		letter's text fields, like substantive SubjectField and BodyField. After obtaining "field_value"
		as a string, will try to apply "modifiers", which must be a dictionary with format-keys strings
		mapped with desired strings. Either "modifiers" are not provided, or successfully applied, returns
		the final string value. For any other case, like invalid or absent "field_value" or if "modifiers"
		failed, None will be returned. This is important to note once again, that if "modifiers" applying
		failed, the original "field_value" string will not be returned.
	"""

	modifiers	:Optional[Dict[str,str]]

	def __call__(self) -> str | None :

		if	(field := super().__call__()) is not None:
			if	isinstance(getattr(self, "modifiers", None), dict):


				try:	final_field = field.format(**self.modifiers)
				except	Exception as E:

					self.loggy.error(f"{self} field failed due to {patronus(E)}")
					return
			else:
				self.loggy.debug("Proper modifiers not found")
				final_field = field


			flen = len(final_field)
			self.loggy.debug(f"Obtained {flen} symbol{flagrate(flen)}")


			return	final_field








class AddressField(ObjectField):

	"""
		hedwig.mail core super class for a email address form of letter fields. Obtained "field_value"
		must be either a single address string, or a bunch of addresses either as a semicolon separated
		string or a sequence of strings (any iterable). In first case, when "field_value" is a string,
		it will be converted to a list of strings by splitting with semicolon. All addresses will be putted
		to a "validator" callable, which is mandatory and must return boolean to decide whether address is
		valid or not. Validated addresses will be gathered and joined as semicolon separated string, which
		is suitable for any recipient like letter fields, but not for sender field, so multiple senders
		must be supplied by multiple fields. Returns joined addresses string in case of at least one
		address was validated. Otherwise, when "field_value" is invalid, or by some Exception that raised
		and logged during iterating over "field_value" or in "validator" call for any address, None will
		be returned. Is a super class for a substantive letter fields like:
			SenderField
			RecipientField
			ccField
			bccField
	"""


	field_value	:str | Sequence[str]
	validator	:EmailValidator


	def __call__(self) -> str | None :


		if	(field := super().__call__()) is not None:
			if	callable(getattr(self, "validator", None)) and not (final_field := list()):


				# At this point callable "validator" is accessible and "field" value might be either
				# a string that content already semicolon separated recipients (or a single recipient),
				# or some other value, which must be an iterable with email addresses as strings. In both
				# situations every email must be validated and then joined back together as final field.
				if	isinstance(field, str): field = ( email.strip() for email in field.split(";") if email )
				try:
					for email in field:
						if	self.validator(email) : final_field.append(email)

				except	Exception as E:

					self.loggy.error(f"{self} field failed due to {patronus(E)}")
				else:
					if	(flen := len(final_field)):

						self.loggy.info(f"{self} obtained {flen} email{flagrate(flen)}")
						return	";".join(final_field) + ";"


					else:	self.loggy.info(f"{self} field has no valid addresses")
			else:			self.loggy.info(f"{self} cannot validate email address")










import	os
from	typing							import List
from	typing							import Dict
from	pygwarts.hedwig.mail.builder	import LetterBuilder
from	pygwarts.magical.spells			import patronus








class ThunderbirdBuilder(LetterBuilder):

	"""
		LetterBuilder implementation of Thunderbird client. Fetches letter fields by LetterBuilder.build
		and invokes cli compose option by os.system for every SenderField entity provided. By default
		Thunderbird client will automatically use default account for sender if provided one not found.
		Thunderbird binary must be included in path environment variable in order to os.system be able
		to call it. Handles no Exceptions that might be raised during os.system call.
	"""

	def build(self) -> Dict[str,str | List[str]] | None :

		"""
			Core method for LetterBuilder implementation of Thunderbird client. Handles no Exceptions.
			Returns fetched by LetterBuilder.build letter dictionary in case of successful sending,
			which is decided by the exit code of compose command, or None otherwise.
		"""

		if	self.is_buildable(current_build := super().build()):

			current_recip	= current_build["recipient"]
			current_cc		= current_build["cc"]
			current_bcc		= current_build["bcc"]
			current_subject	= current_build["subject"]
			current_body	= current_build["body"]
			current_attach	= current_build["attachment"]
			current_call	= f"',to='{current_recip}'"


			if	current_cc is not None		: current_call += f",cc='{current_cc}'"
			if	current_bcc is not None		: current_call += f",bcc='{current_bcc}'"
			if	current_subject is not None : current_call += f",subject='{current_subject}'"
			if	current_body is not None	: current_call += f",body='{current_body}'"
			if	current_attach is not None	: current_call += f",attachment='{','.join(current_attach)}'"


			for sender in current_build["sender"]:
				self.loggy.info(f"{self} commence {sender.rstrip(';')} letter build")

				print(f"\n\n{os.name = }")
				if	os.name == "posix":

					final_call = f"thunderbird -compose \"from='{sender}{current_call}\""
				else:
					final_call = str(

						"C:\\\"Program Files\"\\\"Mozilla Thunderbird\"\\thunderbird.exe"
						f" -compose \"from='{sender}{current_call}\""
					)

				if		not os.system(final_call): self.loggy.info("Letter successfully composed")
				else:	self.loggy.warning("Compose exit code was not 0")
			return		current_build
		else:			self.loggy.info(f"{self} letter build check failed")








class OutlookBuilder(LetterBuilder):

	"""
		LetterBuilder implementation of Outlook client. First of all will try to import win api module and
		to dispatch Outlook application. If ModuleNotFoundError or any other Exception raised during this,
		it is assumed current builder is not applicable for current system. In case of success will fetches
		letter fields by LetterBuilder.build and uses Outlook api to build and explicitly send letter, which
		means there will be no automatic sending, but the constructor window invocation, filled with built
		letter fields. As building goes for every SenderField entity provided, and Outlook restricts sender
		field usage, every SenderField entity will be checked in list of current session accounts beforehand.
		Handles no Exceptions after Outlook application dispatching.
	"""

	def build(self) -> Dict[str,str | List[str]] | None :

		"""
			Core method for LetterBuilder implementation of Outlook client. Handles no Exceptions. As letter
			sending for clients is oriented to letter window construction, return value is fetched by
			LetterBuilder.build letter dictionary in case "build" was handed to a window, which is only
			decided by the api call, or None otherwise.
		"""

		try:

			import win32com.client as win
			app = win.Dispatch("Outlook.Application")

		except	ModuleNotFoundError	: self.loggy.info(f"{self} is not applicable")
		except	Exception as E		: self.loggy.info(f"{self} failed due to {patronus(E)}")
		else:
			if	self.is_buildable(current_build := super().build()):


				self.loggy.debug(f"Found application {app}")
				accounts = { f"{acc};": acc for acc in app.Session.Accounts }


				current_cc		= current_build["cc"]
				current_bcc		= current_build["bcc"]
				current_subject	= current_build["subject"]
				current_body	= current_build["body"]
				current_attach	= current_build["attachment"]


				for sender in current_build["sender"]:
					if	(current_sender := accounts.get(sender)) is not None:


						self.loggy.info(f"{self} commence {sender.rstrip(';')} letter build")


						current_call	= app.CreateItem(0)
						current_call.To = current_build["recipient"]


						if	current_cc is not None		: current_call.CC = current_cc
						if	current_bcc is not None		: current_call.BCC = current_bcc
						if	current_subject is not None	: current_call.Subject = current_subject
						if	current_body is not None	: current_call.HTMLBody = current_body
						if	current_attach is not None	:

							for attachment in current_build["attachment"]:
								current_call.Attachments.Add(attachment)


						current_call._oleobj_.Invoke(64209, 0, 8, 0, current_sender)
						current_call.Display(True)
						self.loggy.info(f"Letter was handed to {app}")


					else:	self.loggy.info(f"Account for {sender.rstrip(';')} not found")
				return		current_build
			else:			self.loggy.info(f"{self} letter build check failed")










import re








MONTHS_NAME_P	= re.compile(

	"([Jj][Aa][Nn]([Uu][Aa][Rr][Yy])?)|"
	"([Ff][Ee][Bb]([Rr][Uu][Aa][Rr][Yy])?)|"
	"([Mm][Aa][Rr]([Cc][Hh])?)|"
	"([Aa][Pp][Rr]([Ii][Ll])?)|"
	"([Mm][Aa][Yy])|"
	"([Jj][Uu][Nn]([Ee])?)|"
	"([Jj][Uu][Ll]([Yy])?)|"
	"([Aa][Uu][Gg]([Uu][Ss][Tt])?)|"
	"([Ss][Ee][Pp]([Tt][Ee][Mm][Bb][Ee][Rr])?)|"
	"([Oo][Cc][Tt]([Oo][Bb][Ee][Rr])?)|"
	"([Nn][Oo][Vv]([Ee][Mm][Bb][Ee][Rr])?)|"
	"([Dd][Ee][Cc]([Ee][Mm][Bb][Ee][Rr])?)"
)




VALID_29_DAYS	= r"(?P<tnd>[12][0-9]|0?[1-9])"
VALID_30_DAYS	= r"(?P<thd>30|[12][0-9]|0?[1-9])"
VALID_31_DAYS	= r"(?P<tod>3[01]|[12][0-9]|0?[1-9])"
VALID_29_MONTHS	= fr"(?P<allM>{MONTHS_NAME_P.pattern}|1[0-2]|0?[1-9])"
VALID_30_MONTHS	= fr"(?P<thdM>{MONTHS_NAME_P.pattern}|11|0?[469])"
VALID_31_MONTHS	= fr"(?P<todM>{MONTHS_NAME_P.pattern}|1[02]|0?[13578])"
VALID_YEAR		= fr"(?P<y>[12]\d\d\d)"
VALID_DELIMITER	= r"[,\\/\.\-\s\:]"
HUNDSCALE_P		= re.compile(r"(?P<base>\d+)?(?P<scale>[0-5]\d)$")




VALID_TIME_P	= re.compile(
	rf"""
		(
			(
				(?P<hh>[0-1]\d|2[0-3]){VALID_DELIMITER}?
			)
			|
			(
				(?P<h>\d){VALID_DELIMITER}
			)
		)
		(
			(?P<mm>[0-5]\d){VALID_DELIMITER}?
			|
			(
				(?P<m>\d)
				(
					{VALID_DELIMITER}(?=\d)
				)?
			)
		)
		(
			(?P<ss>[0-5]\d)|(?P<s>\d)
		)?
	""",
	re.VERBOSE
)




# DD/MM/YYYY
VALID_DATE_1_P	= re.compile(

	f"({VALID_29_DAYS}{VALID_DELIMITER}{VALID_29_MONTHS}"
    f"|{VALID_30_DAYS}{VALID_DELIMITER}{VALID_30_MONTHS}"
    f"|{VALID_31_DAYS}{VALID_DELIMITER}{VALID_31_MONTHS})"
    f"{VALID_DELIMITER}{VALID_YEAR}"
)
DATETIME_1_P	= re.compile(

	fr"(?P<date>{VALID_DATE_1_P.pattern})\ (?P<time>{VALID_TIME_P.pattern})",
	re.VERBOSE
)




# MM/DD/YYYY
VALID_DATE_2_P	= re.compile(

	f"({VALID_29_MONTHS}{VALID_DELIMITER}{VALID_29_DAYS}"
    f"|{VALID_30_MONTHS}{VALID_DELIMITER}{VALID_30_DAYS}"
    f"|{VALID_31_MONTHS}{VALID_DELIMITER}{VALID_31_DAYS})"
    f"{VALID_DELIMITER}{VALID_YEAR}"
)
DATETIME_2_P	= re.compile(

	fr"(?P<date>{VALID_DATE_2_P.pattern})\ (?P<time>{VALID_TIME_P.pattern})",
	re.VERBOSE
)




# YYYY/MM/DD
VALID_DATE_3_P	= re.compile(

    f"{VALID_YEAR}{VALID_DELIMITER}"
	f"({VALID_29_MONTHS}{VALID_DELIMITER}{VALID_29_DAYS}"
    f"|{VALID_30_MONTHS}{VALID_DELIMITER}{VALID_30_DAYS}"
    f"|{VALID_31_MONTHS}{VALID_DELIMITER}{VALID_31_DAYS})"
)
DATETIME_3_P	= re.compile(

	fr"(?P<date>{VALID_DATE_3_P.pattern})\ (?P<time>{VALID_TIME_P.pattern})",
	re.VERBOSE
)




# YYYY/DD/MM
VALID_DATE_4_P	= re.compile(

    f"{VALID_YEAR}{VALID_DELIMITER}"
	f"({VALID_29_DAYS}{VALID_DELIMITER}{VALID_29_MONTHS}"
    f"|{VALID_30_DAYS}{VALID_DELIMITER}{VALID_30_MONTHS}"
    f"|{VALID_31_DAYS}{VALID_DELIMITER}{VALID_31_MONTHS})"
)
DATETIME_4_P	= re.compile(

	fr"(?P<date>{VALID_DATE_4_P.pattern})\ (?P<time>{VALID_TIME_P.pattern})",
	re.VERBOSE
)








def hundscale(sixscale :str | int | float) -> str | Exception :

	"""
		Converts 60 scaled time string to 100 scaled string. Argument "sixscale" might be string,
		integer or float. It must represent the scalable value, which means it's last one or two
		digits have to be in range 0-59. It must be numerical value. The rest, from the beginning,
		will be taken as a base value to concatenate with obtained scaled value. The base value
		might be defaulted to zero for string concatenation. In other words this is conversion
		of hours:minutes or minutes:seconds values that represent time, to fit to the hundred
		scale plot, for example. E.g. 420 become 433, 19 become 031, 6 become 010, 0 become 000,
		e.t.c. Returns concatenated strings with leading zeros sometimes. Raises Exception in case
		of any fail.
	"""

	try:

		scalable = str(int(float(str(sixscale)))).zfill(2)
		BASE, scale = HUNDSCALE_P.fullmatch(scalable).group("base", "scale")

	except	Exception as E : raise ValueError(f"Unscalable argument \"{sixscale}\"")
	else:

		BASE	= BASE or "0"
		SCALE	= str(int(scale)*100 //60).zfill(2)

		return	BASE + SCALE








def monthmap(month	:str) -> str | None :

	""" Mapping english month word with corresponding month number as a string with leading zero """

	if month is not None:

		try:

			parsed_groups	= MONTHS_NAME_P.match(month).groups()
			found_month		= [ i for i,M in enumerate(parsed_groups) if M ][0]


			# After regex been parsed, the very first matched index will point to the month name
			# that matched, so by such indices correspondence, numerical string obtained.
			match found_month:

				case 0	: return "01"
				case 2	: return "02"
				case 4	: return "03"
				case 6	: return "04"
				case 8	: return "05"
				case 9	: return "06"
				case 11	: return "07"
				case 13	: return "08"
				case 15	: return "09"
				case 17	: return "10"
				case 19	: return "11"
				case 21	: return "12"


		# After regex parsing failed, trying to infer month numeric string from the argument.
		except AttributeError:
			match month:

				case "1"	| "01" | 1	: return "01"
				case "2"	| "02" | 2	: return "02"
				case "3"	| "03" | 3	: return "03"
				case "4"	| "04" | 4	: return "04"
				case "5"	| "05" | 5	: return "05"
				case "6"	| "06" | 6	: return "06"
				case "7"	| "07" | 7	: return "07"
				case "8"	| "08" | 8	: return "08"
				case "9"	| "09" | 9	: return "09"
				case "10"	| 10 		: return "10"
				case "11"	| 11 		: return "11"
				case "12"	| 12 		: return "12"










import	re
import	os
import	unittest
from	pathlib								import Path
from	pygwarts.tests.hagrid				import HagridTestCase
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.hagrid.cultivation.sifting	import SiftingController








class SiftingCases(HagridTestCase):

	"""
		Sifting logic
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.SIFT_HANDLER): os.remove(cls.SIFT_HANDLER)

	@classmethod
	def setUpClass(cls): cls.make_loggy_file(cls, cls.SIFT_HANDLER)
	def test_controller_unpatternables(self):

		unpatternables = None, True, 42, .69, SiftingController, [], {}, set(), tuple(),

		class SiftingCase(SiftingController):
			class loggy(LibraryContrib):

				init_level	= 10
				init_name	= "controller_unpatternables"
				handler		= self.SIFT_HANDLER

			include	= unpatternables[:4]
			exclude	= unpatternables[4:]


		with self.assertLogs("controller_unpatternables", 10) as case_loggy:

			self.test_case = SiftingCase()
			for candidate in unpatternables:

				self.assertIn(
					f"WARNING:controller_unpatternables:Unpatternable type {type(candidate)}",
					case_loggy.output
				)
			self.assertFalse(hasattr(self.test_case, "exclude_field"))
			self.assertFalse(hasattr(self.test_case, "include_field"))




	def test_controller_wrong(self):

		unpatternables = None, True, 42, .69, SiftingController, [], {}, set(), tuple(),

		class SiftingCase(SiftingController):
			class loggy(LibraryContrib):

				init_level	= 10
				init_name	= "controller_wrongs"
				handler		= self.SIFT_HANDLER

			include	= tuple()
			exclude	= "tuple()"


		with self.assertLogs("controller_wrongs", 10) as case_loggy:

			self.test_case = SiftingCase()
			for candidate in unpatternables:

				self.assertIn(
					"WARNING:controller_wrongs:"
					f"{self.test_case} exclude field improper, must be not empty list or tuple",
					case_loggy.output
				)
				self.assertIn(
					"WARNING:controller_wrongs:"
					f"{self.test_case} include field improper, must be not empty list or tuple",
					case_loggy.output
				)
			self.assertFalse(hasattr(self.test_case, "exclude_field"))
			self.assertFalse(hasattr(self.test_case, "include_field"))




	def test_controller_sprigs(self):
		class SiftingCase(SiftingController):
			class loggy(LibraryContrib):

				init_level	= 10
				init_name	= "controller_sprigs"
				handler		= self.SIFT_HANDLER


		self.test_case = SiftingCase()
		invalid_types = "sprigs", ( "sprigs", ), { "sprigs" }, iter([ "sprigs" ]),
		invalid_sprigs = [ "sprigs" ], [( "sprigs", )], [{ "sprigs" }], [ iter([ "sprigs" ])],


		for item1 in invalid_types:
			with self.assertLogs("controller_sprigs", 10) as case_loggy:

				self.test_case(item1)
				self.assertIn(
					f"WARNING:controller_sprigs:{self.test_case} found invalid sprigs type {type(item1)}",
					case_loggy.output
				)


		for item2 in invalid_sprigs:
			with self.assertLogs("controller_sprigs", 10) as case_loggy:

				self.test_case(item2)
				self.assertIn(
					f"WARNING:controller_sprigs:{self.test_case} found invalid sprig \"{item2[0]}\"",
					case_loggy.output
				)




	def test_controller_no_Sieve(self):
		class SiftingCase(SiftingController):
			class loggy(LibraryContrib):

				init_level	= 10
				init_name	= "controller_no_Sieve"
				handler		= self.SIFT_HANDLER


		self.test_case = SiftingCase()
		self.test_case.Sieve = "Sieve"


		with self.assertLogs("controller_no_Sieve", 10) as case_loggy:

			self.test_case("sprigs")
			self.assertIn(
				"DEBUG:controller_no_Sieve:Sieve not implemented for current controller",
				case_loggy.output
			)




	def test_controller_fields(self):
		class SiftingCase(SiftingController):
			class loggy(LibraryContrib):

				init_level	= 10
				init_name	= "controller_fields"
				handler		= self.SIFT_HANDLER


		with self.assertLogs("controller_fields", 10) as case_loggy:

			self.test_case = SiftingCase()
			self.assertIn(
				"DEBUG:controller_fields:Exclude field not implemented for sifting", case_loggy.output
			)
			self.assertIn(
				"DEBUG:controller_fields:Exclude field not implemented for sifting", case_loggy.output
			)
			self.assertFalse(hasattr(self.test_case, "exclude_field"))
			self.assertFalse(hasattr(self.test_case, "include_field"))




		with self.assertLogs("controller_fields", 10) as case_loggy:

			self.test_case([ Path("False") ])
			self.assertIn(
				"DEBUG:controller_fields:Exclude field skipped", case_loggy.output
			)
			self.assertIn(
				"DEBUG:controller_fields:Exclude field skipped", case_loggy.output
			)
			self.assertFalse(hasattr(self.test_case, "exclude_field"))
			self.assertFalse(hasattr(self.test_case, "include_field"))




		with self.assertLogs("controller_fields", 10) as case_loggy:

			self.test_case.include_field = []
			self.test_case([ Path("False") ])
			self.assertIn(
				"DEBUG:controller_fields:Exclude field skipped", case_loggy.output
			)
			self.assertIn(
				"DEBUG:controller_fields:Sifted sprig \"False\"", case_loggy.output
			)
			self.assertFalse(hasattr(self.test_case, "exclude_field"))
			self.assertTrue(hasattr(self.test_case, "include_field"))




		with self.assertLogs("controller_fields", 10) as case_loggy:

			self.test_case.exclude_field = []
			self.test_case([ Path("False") ])
			self.assertIn(
				"DEBUG:controller_fields:Exclude field passed", case_loggy.output
			)
			self.assertIn(
				"DEBUG:controller_fields:Sifted sprig \"False\"", case_loggy.output
			)
			self.assertTrue(hasattr(self.test_case, "exclude_field"))
			self.assertTrue(hasattr(self.test_case, "include_field"))




		with self.assertLogs("controller_fields", 10) as case_loggy:

			self.test_case.include_field = [ re.compile("False") ]
			self.test_case([ Path("False") ])
			self.assertIn(
				"DEBUG:controller_fields:Exclude field passed", case_loggy.output
			)
			self.assertIn(
				"DEBUG:controller_fields:Passed sprig \"False\"", case_loggy.output
			)
			self.assertTrue(hasattr(self.test_case, "exclude_field"))
			self.assertTrue(hasattr(self.test_case, "include_field"))




		with self.assertLogs("controller_fields", 10) as case_loggy:

			self.test_case.exclude_field = [ re.compile("False") ]
			self.test_case([ Path("False") ])
			self.assertIn(
				"DEBUG:controller_fields:Sifted sprig \"False\"", case_loggy.output
			)
			self.assertTrue(hasattr(self.test_case, "exclude_field"))
			self.assertTrue(hasattr(self.test_case, "include_field"))




		with self.assertLogs("controller_fields", 10) as case_loggy:

			self.test_case.exclude_field = self.test_case.pmake_field([])
			self.assertIn(
				"DEBUG:controller_fields:Making field from 0 entities", case_loggy.output
			)

			self.test_case.include_field = self.test_case.pmake_field([])
			self.assertIn(
				"DEBUG:controller_fields:Making field from 0 entities", case_loggy.output
			)

			self.assertFalse(self.test_case.exclude_field)
			self.assertFalse(self.test_case.include_field)

			self.test_case.include_field = self.test_case.pmake_field([ "False" ])
			self.assertTrue(self.test_case.include_field)
			self.assertIn(
				"DEBUG:controller_fields:Making field from 1 entities", case_loggy.output
			)

			self.test_case([ Path("False") ])
			self.assertIn(
				"DEBUG:controller_fields:Passed sprig \"False\"", case_loggy.output
			)




		with self.assertLogs("controller_fields", 10) as case_loggy:

			self.test_case.exclude_field = 42
			self.test_case.include_field = 42

			self.assertEqual(self.test_case.exclude_field, 42)
			self.assertEqual(self.test_case.include_field, 42)

			self.test_case([ Path("False") ], excludables=[], includables=[])
			self.assertIn(
				"DEBUG:controller_fields:Sifted sprig \"False\"", case_loggy.output
			)




		with self.assertLogs("controller_fields", 10) as case_loggy:

			self.test_case.exclude_field = 420
			self.test_case.include_field = 420

			self.assertEqual(self.test_case.exclude_field, 420)
			self.assertEqual(self.test_case.include_field, 420)

			self.test_case([ Path("False") ], excludables=[], includables=[ re.compile("False") ])
			self.assertIn(
				"DEBUG:controller_fields:Passed sprig \"False\"", case_loggy.output
			)




		with self.assertLogs("controller_fields", 10) as case_loggy:

			self.test_case.exclude_field = 69
			self.test_case.include_field = 69

			self.assertEqual(self.test_case.exclude_field, 69)
			self.assertEqual(self.test_case.include_field, 69)

			self.test_case(

				[ Path("False") ],
				excludables=[ re.compile("False") ],
				includables=[ re.compile("False") ])
			self.assertIn(
				"DEBUG:controller_fields:Sifted sprig \"False\"", case_loggy.output
			)








	def test_Sieve_siftable(self):
		class SiftingCase(SiftingController):
			class loggy(LibraryContrib):

				init_level	= 10
				init_name	= "Sieve_siftable"
				handler		= self.SIFT_HANDLER


		self.test_case = SiftingCase()


		with self.assertLogs("Sieve_siftable", 10) as case_loggy:

			sifted = self.test_case.Sieve(False, exclude=None, include=None, thriving=False)
			self.assertIn(
				"DEBUG:Sieve_siftable:Got invalid siftable \"False\"", case_loggy.output
			)
			self.assertTrue(sifted)




	def test_Sieve_empties(self):
		class SiftingCase(SiftingController):
			class loggy(LibraryContrib):

				init_level	= 10
				init_name	= "Sieve_empties"
				handler		= self.SIFT_HANDLER


		self.test_case = SiftingCase()


		with self.assertLogs("Sieve_empties", 10) as case_loggy:

			sifted = self.test_case.Sieve("False", exclude=[], include=[], thriving=False)
			self.assertIn(
				"DEBUG:Sieve_empties:Exclude field passed", case_loggy.output
			)
			self.assertIn(
				"DEBUG:Sieve_empties:Include field not passed", case_loggy.output
			)
			self.assertTrue(sifted)




	def test_Sieve_Nones(self):
		class SiftingCase(SiftingController):
			class loggy(LibraryContrib):

				init_level	= 10
				init_name	= "Sieve_Nones"
				handler		= self.SIFT_HANDLER


		self.test_case = SiftingCase()


		with self.assertLogs("Sieve_Nones", 10) as case_loggy:

			sifted = self.test_case.Sieve("False", exclude=None, include=None, thriving=False)
			self.assertIn(
				"DEBUG:Sieve_Nones:Exclude field skipped", case_loggy.output
			)
			self.assertIn(
				"DEBUG:Sieve_Nones:Include field skipped", case_loggy.output
			)
			self.assertFalse(sifted)




	def test_Sieve_fields(self):
		class SiftingCase(SiftingController):
			class loggy(LibraryContrib):

				init_level	= 10
				init_name	= "Sieve_fields"
				handler		= self.SIFT_HANDLER

			exclude_field	= "excludable",
			include_field	= "includable",


		self.test_case = SiftingCase()


		with self.assertLogs("Sieve_fields", 10) as case_loggy:

			sifted = self.test_case.Sieve(

				"False",
				exclude=self.test_case.exclude_field,
				include=self.test_case.include_field,
				thriving=False
			)
			self.assertIn(
				f"DEBUG:Sieve_fields:Got invalid exclude field", case_loggy.output
			)
			self.assertIn(
				f"DEBUG:Sieve_fields:Got invalid include field", case_loggy.output
			)
			self.assertTrue(sifted)




	def test_Sieve_invalids(self):
		class SiftingCase(SiftingController):
			class loggy(LibraryContrib):

				init_level	= 10
				init_name	= "Sieve_invalids"
				handler		= self.SIFT_HANDLER

			exclude_field	= [ "not Path" ]
			include_field	= [ "not Path" ]


		self.test_case = SiftingCase()


		with self.assertLogs("Sieve_invalids", 10) as case_loggy:

			sifted = self.test_case.Sieve(

				"False",
				exclude=self.test_case.exclude_field,
				include=self.test_case.include_field,
				thriving=False
			)
			self.assertIn(

				f"DEBUG:Sieve_invalids:Got invalid excludable \"not Path\"",
				case_loggy.output
			)
			self.assertIn(
				"DEBUG:Sieve_invalids:Got invalid includable \"not Path\"", case_loggy.output
			)
			self.assertTrue(sifted)
















	def test_fs_strings(self):
		if	os.name == "posix":

			class SiftingCase(SiftingController):
				class loggy(LibraryContrib):

					init_level	= 10
					init_name	= "controller_strings"
					handler		= self.SIFT_HANDLER

				include	= (

					r".+\.mp[34]",
					rf".+/good/[^/]+",	# The end folder is good for any file
				)
				exclude	= rf".+scooter.+",


			self.test_case = SiftingCase()
			current_case = {

				"/home/user/music/prodigy":		[

					"thier law.mp3", "thier law.mp4", "thier law.mp5",
					"baby's got a temper.mp3", "baby's got a temper.mp4", "baby's got a temper.mp5",
					"breath.mp3", "breath.mp4", "breath.mp5"
				],
				"/home/user/music/scooter":		[ "how much is the fish.mp3" ],
				"/srv/web/cache":				[ "not.trojan" ],
				"/srv/storage/memories/good":	[ "day1.jpg", "day2.jpg" ],
				"/srv/storage/days/good/photos":[ "day1.jpg", "day2.jpg" ],
			}


			with self.assertLogs("controller_strings", 10) as case_loggy:
				self.assertEqual(

					[
						p.name for p in self.test_case(
							[
								Path(f"/home/user/music/prodigy/{item}")
								for item in current_case["/home/user/music/prodigy"]
							]
						)
					],	[
							"thier law.mp3", "thier law.mp4",
							"baby's got a temper.mp3", "baby's got a temper.mp4",
							"breath.mp3", "breath.mp4",
						]
				)
				self.assertIn(
					"DEBUG:controller_strings:Passed sprig \"thier law.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_strings:Passed sprig \"thier law.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_strings:Sifted sprig \"thier law.mp5\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_strings:Passed sprig \"baby's got a temper.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_strings:Passed sprig \"baby's got a temper.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_strings:Sifted sprig \"baby's got a temper.mp5\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_strings:Passed sprig \"breath.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_strings:Passed sprig \"breath.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_strings:Sifted sprig \"breath.mp5\"", case_loggy.output
				)


			with self.assertLogs("controller_strings", 10) as case_loggy:
				self.assertEqual(
					self.test_case(
						[
							Path(f"/home/user/music/scooter/{item}") for item in
							current_case["/home/user/music/scooter"]
						]
					),	[]
				)
				self.assertIn(
					"DEBUG:controller_strings:Sifted sprig \"how much is the fish.mp3\"", case_loggy.output
				)


			with self.assertLogs("controller_strings", 10) as case_loggy:
				self.assertEqual(
					self.test_case(
						[ Path(f"/srv/web/cache/{item}") for item in current_case["/srv/web/cache"] ]
					),	[]
				)
				self.assertIn(
					"DEBUG:controller_strings:Sifted sprig \"not.trojan\"", case_loggy.output
				)


			with self.assertLogs("controller_strings", 10) as case_loggy:
				self.assertEqual(
					[
						p.name for p in self.test_case(
							[
								Path(f"/srv/storage/memories/good/{item}") for item in
								current_case["/srv/storage/memories/good"]
							]
						)
					],	[ "day1.jpg", "day2.jpg" ]
				)
				self.assertIn(
					"DEBUG:controller_strings:Passed sprig \"day1.jpg\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_strings:Passed sprig \"day2.jpg\"", case_loggy.output
				)


			with self.assertLogs("controller_strings", 10) as case_loggy:
				self.assertEqual(
					self.test_case(
						[
							Path(f"/srv/storage/days/good/photos/{item}") for item in
							current_case["/srv/storage/days/good/photos"]
						]
					),	[]
				)
				self.assertIn(
					"DEBUG:controller_strings:Sifted sprig \"day1.jpg\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_strings:Sifted sprig \"day2.jpg\"", case_loggy.output
				)




		elif(os.name == "nt"):

			class SiftingCase(SiftingController):
				class loggy(LibraryContrib):

					init_level	= 10
					init_name	= "otherfs_controller_strings"
					handler		= self.SIFT_HANDLER

				include	= (

					r".+\.mp[34]",
					rf".+\\good\\[^(\\)]+",	# The end folder is good for any file
				)
				exclude	= rf".+scooter.+",


			self.test_case = SiftingCase()
			current_case = {

				"D:\\music\\prodigy":				[

					"thier law.mp3", "thier law.mp4", "thier law.mp5",
					"baby's got a temper.mp3", "baby's got a temper.mp4", "baby's got a temper.mp5",
					"breath.mp3", "breath.mp4", "breath.mp5"
				],
				"D:\\music\\scooter":				[ "how much is the fish.mp3" ],
				"C:\\Program Files":				[ "not.trojan" ],
				"E:\\storage\\memories\\good":		[ "day1.jpg", "day2.jpg" ],
				"F:\\storage\\days\\good\\photos":	[ "day1.jpg", "day2.jpg" ],
			}


			with self.assertLogs("otherfs_controller_strings", 10) as case_loggy:
				self.assertEqual(

					[
						p.name for p in self.test_case(
							[
								Path(f"D:\\music\\prodigy\\{item}")
								for item in  current_case["D:\\music\\prodigy"]
							]
						)
					],	[
						"thier law.mp3", "thier law.mp4",
						"baby's got a temper.mp3", "baby's got a temper.mp4",
						"breath.mp3", "breath.mp4",
						]
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Passed sprig \"thier law.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Passed sprig \"thier law.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Sifted sprig \"thier law.mp5\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Passed sprig \"baby's got a temper.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Passed sprig \"baby's got a temper.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Sifted sprig \"baby's got a temper.mp5\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Passed sprig \"breath.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Passed sprig \"breath.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Sifted sprig \"breath.mp5\"", case_loggy.output
				)


			with self.assertLogs("otherfs_controller_strings", 10) as case_loggy:
				self.assertEqual(
					self.test_case(

						[ Path(f"D:\\music\\scooter\\{item}") for item in current_case["D:\\music\\scooter"] ]
					),	[]
				)
				self.assertIn(

					"DEBUG:otherfs_controller_strings:Sifted sprig \"how much is the fish.mp3\"",
					case_loggy.output
				)


			with self.assertLogs("otherfs_controller_strings", 10) as case_loggy:
				self.assertEqual(
					self.test_case(
						[ Path(f"C:\\Program Files\\{item}") for item in current_case["C:\\Program Files"] ]
					),	[]
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Sifted sprig \"not.trojan\"", case_loggy.output
				)


			with self.assertLogs("otherfs_controller_strings", 10) as case_loggy:
				self.assertEqual(
					[
						p.name for p in self.test_case(
							[
								Path(f"E:\\storage\\memories\\good\\{item}") for item in
								current_case["E:\\storage\\memories\\good"]
							]
						)
					],	[ "day1.jpg", "day2.jpg" ]
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Passed sprig \"day1.jpg\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Passed sprig \"day2.jpg\"", case_loggy.output
				)


			with self.assertLogs("otherfs_controller_strings", 10) as case_loggy:
				self.assertEqual(
					self.test_case(
						[
							Path(f"F:\\storage\\days\\good\\photos\\{item}") for item in
							current_case["F:\\storage\\days\\good\\photos"]
						]
					),	[]
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Sifted sprig \"day1.jpg\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_strings:Sifted sprig \"day2.jpg\"", case_loggy.output
				)
		else:	self.assertFalse("OS neither posix nor windows")








	def test_fs_Patterns(self):
		if	os.name == "posix":

			class SiftingCase(SiftingController):
				class loggy(LibraryContrib):

					init_level	= 10
					init_name	= "controller_Patterns"
					handler		= self.SIFT_HANDLER

				include	= (

					re.compile(r".+\.mp[34]"),
					re.compile(rf".+/good/[^/]+"),	# The end folder is good for any file
				)
				exclude	= re.compile(rf".+scooter.+"),


			self.test_case = SiftingCase()
			current_case = {

				"/home/user/music/prodigy":		[

					"thier law.mp3", "thier law.mp4", "thier law.mp5",
					"baby's got a temper.mp3", "baby's got a temper.mp4", "baby's got a temper.mp5",
					"breath.mp3", "breath.mp4", "breath.mp5"
				],
				"/home/user/music/scooter":		[ "how much is the fish.mp3" ],
				"/srv/web/cache":				[ "not.trojan" ],
				"/srv/storage/memories/good":	[ "day1.jpg", "day2.jpg" ],
				"/srv/storage/days/good/photos":[ "day1.jpg", "day2.jpg" ],
			}


			with self.assertLogs("controller_Patterns", 10) as case_loggy:
				self.assertEqual(

					[
						p.name for p in self.test_case(
							[
								Path(f"/home/user/music/prodigy/{item}")
								for item in current_case["/home/user/music/prodigy"]
							]
						)
					],	[
							"thier law.mp3", "thier law.mp4",
							"baby's got a temper.mp3", "baby's got a temper.mp4",
							"breath.mp3", "breath.mp4",
						]
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Passed sprig \"thier law.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Passed sprig \"thier law.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Sifted sprig \"thier law.mp5\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Passed sprig \"baby's got a temper.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Passed sprig \"baby's got a temper.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Sifted sprig \"baby's got a temper.mp5\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Passed sprig \"breath.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Passed sprig \"breath.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Sifted sprig \"breath.mp5\"", case_loggy.output
				)


			with self.assertLogs("controller_Patterns", 10) as case_loggy:
				self.assertEqual(
					self.test_case(
						[
							Path(f"/home/user/music/scooter/{item}") for item in
							current_case["/home/user/music/scooter"]
						]
					),	[]
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Sifted sprig \"how much is the fish.mp3\"", case_loggy.output
				)


			with self.assertLogs("controller_Patterns", 10) as case_loggy:
				self.assertEqual(
					self.test_case(
						[ Path(f"/srv/web/cache/{item}") for item in current_case["/srv/web/cache"] ]
					),	[]
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Sifted sprig \"not.trojan\"", case_loggy.output
				)


			with self.assertLogs("controller_Patterns", 10) as case_loggy:
				self.assertEqual(
					[
						p.name for p in self.test_case(
							[
								Path(f"/srv/storage/memories/good/{item}") for item in
								current_case["/srv/storage/memories/good"]
							]
						)
					],	[ "day1.jpg", "day2.jpg" ]
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Passed sprig \"day1.jpg\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Passed sprig \"day2.jpg\"", case_loggy.output
				)


			with self.assertLogs("controller_Patterns", 10) as case_loggy:
				self.assertEqual(
					self.test_case(
						[
							Path(f"/srv/storage/days/good/photos/{item}") for item in
							current_case["/srv/storage/days/good/photos"]
						]
					),	[]
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Sifted sprig \"day1.jpg\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:controller_Patterns:Sifted sprig \"day2.jpg\"", case_loggy.output
				)




		elif(os.name == "nt"):

			class SiftingCase(SiftingController):
				class loggy(LibraryContrib):

					init_level	= 10
					init_name	= "otherfs_controller_Patterns"
					handler		= self.SIFT_HANDLER

				include	= (

					re.compile(r".+\.mp[34]"),
					re.compile(rf".+\\good\\[^(\\)]+"),	# The end folder is good for any file
				)
				exclude	= re.compile(rf".+scooter.+"),


			self.test_case = SiftingCase()
			current_case = {

				"D:\\music\\prodigy":				[

					"thier law.mp3", "thier law.mp4", "thier law.mp5",
					"baby's got a temper.mp3", "baby's got a temper.mp4", "baby's got a temper.mp5",
					"breath.mp3", "breath.mp4", "breath.mp5"
				],
				"D:\\music\\scooter":				[ "how much is the fish.mp3" ],
				"C:\\Program Files":				[ "not.trojan" ],
				"E:\\storage\\memories\\good":		[ "day1.jpg", "day2.jpg" ],
				"F:\\storage\\days\\good\\photos":	[ "day1.jpg", "day2.jpg" ],
			}


			with self.assertLogs("otherfs_controller_Patterns", 10) as case_loggy:
				self.assertEqual(

					[
						p.name for p in self.test_case(
							[
								Path(f"D:\\music\\prodigy\\{item}")
								for item in  current_case["D:\\music\\prodigy"]
							]
						)
					],	[
						"thier law.mp3", "thier law.mp4",
						"baby's got a temper.mp3", "baby's got a temper.mp4",
						"breath.mp3", "breath.mp4",
						]
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Passed sprig \"thier law.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Passed sprig \"thier law.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Sifted sprig \"thier law.mp5\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Passed sprig \"baby's got a temper.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Passed sprig \"baby's got a temper.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Sifted sprig \"baby's got a temper.mp5\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Passed sprig \"breath.mp3\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Passed sprig \"breath.mp4\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Sifted sprig \"breath.mp5\"", case_loggy.output
				)


			with self.assertLogs("otherfs_controller_Patterns", 10) as case_loggy:
				self.assertEqual(
					self.test_case(
						[ Path(f"D:\\music\\scooter\\{item}") for item in current_case["D:\\music\\scooter"] ]
					),	[]
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Sifted sprig \"how much is the fish.mp3\"", case_loggy.output
				)


			with self.assertLogs("otherfs_controller_Patterns", 10) as case_loggy:
				self.assertEqual(
					self.test_case(
						[ Path(f"C:\\Program Files\\{item}") for item in current_case["C:\\Program Files"] ]
					),	[]
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Sifted sprig \"not.trojan\"", case_loggy.output
				)


			with self.assertLogs("otherfs_controller_Patterns", 10) as case_loggy:
				self.assertEqual(
					[
						p.name for p in self.test_case(
							[
								Path(f"E:\\storage\\memories\\good\\{item}") for item in
								current_case["E:\\storage\\memories\\good"]
							]
						)
					],	[ "day1.jpg", "day2.jpg" ]
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Passed sprig \"day1.jpg\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Passed sprig \"day2.jpg\"", case_loggy.output
				)


			with self.assertLogs("otherfs_controller_Patterns", 10) as case_loggy:
				self.assertEqual(
					self.test_case(
						[
							Path(f"F:\\storage\\days\\good\\photos\\{item}") for item in
							current_case["F:\\storage\\days\\good\\photos"]
						]
					),	[]
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Sifted sprig \"day1.jpg\"", case_loggy.output
				)
				self.assertIn(
					"DEBUG:otherfs_controller_Patterns:Sifted sprig \"day2.jpg\"", case_loggy.output
				)
		else:	self.assertFalse("OS neither posix nor windows")








if	__name__ == "__main__" : unittest.main(verbosity=2)










from pathlib								import Path
from shutil									import rmtree
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.hagrid.thrivables				import Tree








class TrimProbe(Transmutable):

	"""
		hagrid.planting object, that represents simulation of a folder/file removing operation.
		The purpose of this object is to indicate the intention of real operations, before they will really
		affect the file system.
		This object's __call__ accepts arguments that implement the "planting interface":
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			sprig	- Path object that represents current operation target folder/file;
			bough	- Path object that represents current operation destination parent folder.
		There are three variants for this object __call__ to do:
			1. "sprig" is a folder and corresponding probe message will be logged;
			2. "sprig" is a file and corresponding probe message will be logged;
			1. "sprig" neither a folder nor file and corresponding probe message will be logged.
	"""

	def __call__(self, origin :Tree, sprout :str, branch :Path, sprig :Path, bough :Path):

		if	not sprig.is_symlink():

			if	sprig.is_dir()		: origin.loggy.info(f"Twig \"{sprig}\" trim probe")
			elif(sprig.is_file())	: origin.loggy.info(f"Leaf \"{sprig}\" trim probe")
			else					: origin.loggy.warning(f"Bad sprig \"{sprig}\" to probe")
		else						: self.loggy.debug("Symbolic sprig skipped")








class SprigTrimmer(Transmutable):

	"""
		hagrid.planting object, that represents folder/file removing operation.
		This object's __call__ accepts arguments that implement the "planting interface":
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			sprig	- Path object that represents current operation target folder/file;
			bough	- Path object that represents current operation destination parent folder.
		As sprig can be either destination folder or file, it's type will be determined by __call__. Will be
		used shutil.rmtree for folders and Path.unlink method for files. If any other "sprig" time occurred,
		only corresponding warning message will be logged.
		No return value for __call__ is assumed, also no handling for any Exception, that might be raised
		by "rmtree" or by "Path.unlink", that must be handled by invoker object, e.g. Bloom object.
	"""

	def __call__(self, origin :Tree, sprout :str, branch :Path, sprig :Path, bough :Path):


		if	not sprig.is_symlink():
			if	sprig.is_dir():


				rmtree(sprig)
				origin.loggy.info(f"Trimmed twig \"{sprig}\"")


			elif(sprig.is_file()):

				sprig.unlink()
				origin.loggy.info(f"Trimmed leaf \"{sprig}\"")
			else:
				origin.loggy.warning(f"Bad sprig \"{sprig}\" to trim")
		else:	self.loggy.debug("Symbolic sprig skipped")







